package lu.uni.mobile.bonk;

import java.util.LinkedList;

import lu.uni.mobile.bonk.GameObject.Bullet;
import android.animation.Animator.AnimatorListener;
import android.animation.Animator;
import android.animation.AnimatorSet;
import android.animation.ObjectAnimator;
import android.util.Log;
import android.view.animation.LinearInterpolator;
import android.widget.ImageView;

public abstract class BonkAnimator {
    protected PlaygroundView playgroundView;
    protected AnimatorSet animatorSet;
    protected Game game;

    public BonkAnimator(PlaygroundView playgroundView, Game game) {
        this.playgroundView = playgroundView;
        this.game = game;
    }

    public abstract void animate();

    public void stop() {
        animatorSet.cancel();
    };

    public static class Trajectory extends BonkAnimator {
        protected Bullet bullet;

        public Trajectory(PlaygroundView playgroundView, Bullet bullet, Game game) {
            super(playgroundView, game);
            this.bullet = bullet;
        }

        public void animate() {
            // TODO create tick listener for bullet-only collision detection
            double duration = bullet.getRemainingLifeMS();
            double distance = duration * bullet.getSpeedPerMS();

            double rotationCcwDegrees = bullet.getRotationCcwDegrees();
            double rotationCcwRadians = Math.toRadians(rotationCcwDegrees);
            double dX = Math.cos(rotationCcwRadians) * distance;
            double dY = Math.sin(rotationCcwRadians) * distance;

            animatorSet = new AnimatorSet();
            animatorSet.setInterpolator(new LinearInterpolator());

            // make the bullet disappear at the end of its travels
            animatorSet.addListener(new DisappearWhenFinishedAnimationListener(bullet));

            createAnimationSegmentsForAxis("x", bullet.getXDouble(), dX, duration, null);
            createAnimationSegmentsForAxis("y", bullet.getYDouble(), dY, duration, null);
            animatorSet.start();
        }

        private void createAnimationSegmentsForAxis(String axis, double currentCoordinate,
                double totalRemainingOffset, double totalRemainingDuration, ObjectAnimator previousAnimator) {

            double thisSegmentDuration = totalRemainingDuration;
            double finalCoordinate = currentCoordinate + totalRemainingOffset;
            boolean shouldRecurse = true;
            boolean gameObjectIsOwnedByThisUser = (game.getThisPlayerId() == Game.getObjectOwnerId(bullet.getGameObjectId()));

            if (finalCoordinate > Game.MAX_COORDINATE) {
                // the bullet hit the right or bottom wall
                thisSegmentDuration = totalRemainingDuration * (Game.MAX_COORDINATE - currentCoordinate)
                        / totalRemainingOffset;
                totalRemainingOffset = Game.MAX_COORDINATE - finalCoordinate; // needs to be negative for next
                // animation segment
                finalCoordinate = Game.MAX_COORDINATE;
            }
            else if (finalCoordinate < Game.MIN_COORDINATE) {
                // the bullet hit the left or top wall
                thisSegmentDuration = totalRemainingDuration * currentCoordinate / -totalRemainingOffset;
                totalRemainingOffset = -finalCoordinate; // will be positive for next animation segment
                finalCoordinate = 0.0;
            }
            else {
                // this will be the last segment, because it ends before the next impact.
                shouldRecurse = false;
            }

            ObjectAnimator animator = getObjectAnimatorForTrajectory(
                axis,
                (long) thisSegmentDuration,
                finalCoordinate);

            Log.i("ValueAnimator", String.format(
                "created %s animator from %.2f to %.2f taking %.0fms.",
                axis,
                currentCoordinate,
                finalCoordinate,
                thisSegmentDuration));

            if (previousAnimator != null) {
                // the common case: multiple animation segments being chained together
                animatorSet.play(animator).after(previousAnimator);
                // wall impact only happens between animators, not on the first one.
                animator.addListener(new AxisImpactAnimatorListener(
                    this,
                    bullet,
                    axis,
                    gameObjectIsOwnedByThisUser));
            }
            else {
                // this is the first animation in the series.
                // might be just a single-segment animation, so just add it. If more are chained together via
                // recursion, then they will be added with .play().after(), in the above block.
                animatorSet.play(animator);

                // check this segment for explosions.
                checkCurrentTrajectoryForImpactEvents(animator);
            }

            if (shouldRecurse) {
                createAnimationSegmentsForAxis(
                    axis,
                    finalCoordinate,
                    totalRemainingOffset,
                    totalRemainingDuration - thisSegmentDuration,
                    animator);
            }
        }

        public ObjectAnimator getObjectAnimatorForTrajectory(String axis, long thisSegmentDuration,
                double finalCoordinate) {
            ObjectAnimator animator = ObjectAnimator.ofFloat(bullet, axis, bullet.getPixel(
                axis,
                finalCoordinate));
            animator.setInterpolator(new LinearInterpolator());
            animator.setDuration(thisSegmentDuration);
            return animator;
        }

        private static final double IMPACT_DISTANCE_THRESHOLD = 0.15; // in screen proportion 0.0-1.0

        /**
         * do the necessary geometry to know if this flight segment will hit any fixed objects
         * 
         * @param animation
         * @return
         */
        private void checkCurrentTrajectoryForImpactEvents(Animator animation) {
            // 0: start by assuming there will not be any impact
            long earliestDurationUntilImpact = Long.MAX_VALUE;
            GameObject impactingObject = null;

            // 1. calculate y = mx + b
            double y = bullet.getYDouble();
            double x = bullet.getXDouble();
            double theta = Math.toRadians(bullet.getRotation());

            // for each object:
            for (GameObject otherObject : game.getGameObjects().values()) {
                if (otherObject.isMovable()) {
                    continue; // it's really, really hard to calculate moving impacts
                }
                if (otherObject.getGameObjectId() == game.getThisPlayerId()) {
                    continue;// don't let the bullets blow up their own players
                }

                // location of the other object
                double yPrime = otherObject.getYDouble();
                double xPrime = otherObject.getXDouble();

                // find the difference in angles between trajectory and the other object
                double thetaPrime = Math.atan2(yPrime - y, xPrime - x);
                double dTheta = Math.abs(thetaPrime - theta);

                // sanity check: if the other object is > 90 degrees from the trajectory, then skip this
                if (dTheta > Math.PI / 2) {
                    continue;
                }

                // find the distance at the closest point
                double distanceToTarget = Math.hypot(xPrime - x, yPrime - y);
                double distanceBetweenObjects = distanceToTarget * Math.tan(dTheta);

                // if distance < threshold:
                if (distanceBetweenObjects < IMPACT_DISTANCE_THRESHOLD) {
                    // create an impact event (add it as an animation at a certain time)
                    long durationUntilThisObjectImpact = (long) (distanceToTarget / bullet.getSpeedPerMS());
                    // set the trajectory to end at the EARLIEST impact
                    if (durationUntilThisObjectImpact < earliestDurationUntilImpact) {
                        earliestDurationUntilImpact = durationUntilThisObjectImpact;
                        impactingObject = otherObject;
                    }
                }
            }

            if (impactingObject != null) {
                // this animation segment will result in an impact!
                // replace all animations with a straight-shot animation.
                animatorSet.removeAllListeners();
                animatorSet.cancel();
                AnimatorSet explosionTrajectoryAnimatorSet = new AnimatorSet();

                final ImageView explosionImage = new ImageView(playgroundView.getContext());
                explosionImage.setImageResource(R.drawable.explosion);
                explosionImage.setX(impactingObject.getX());
                explosionImage.setY(impactingObject.getY());
                explosionImage.setScaleX(0.0f);
                explosionImage.setScaleY(0.0f);
                playgroundView.addView(explosionImage);

                ObjectAnimator xTrajectory = getObjectAnimatorForTrajectory(
                    "x",
                    earliestDurationUntilImpact,
                    impactingObject.getX());
                ObjectAnimator yTrajectory = getObjectAnimatorForTrajectory(
                    "y",
                    earliestDurationUntilImpact,
                    impactingObject.getY());
                explosionTrajectoryAnimatorSet.play(xTrajectory);
                explosionTrajectoryAnimatorSet.play(yTrajectory);
                explosionTrajectoryAnimatorSet.play(getExplosionAnimator("scaleX", explosionImage)).after(
                    xTrajectory);
                explosionTrajectoryAnimatorSet.play(getExplosionAnimator("scaleY", explosionImage)).after(
                    yTrajectory);
                explosionTrajectoryAnimatorSet.setDuration(earliestDurationUntilImpact);
                explosionTrajectoryAnimatorSet.addListener(new DisappearWhenFinishedAnimationListener(
                    explosionImage));
                explosionTrajectoryAnimatorSet.start();
            }
        }

        private static final float MAXIMUM_XY_SCALE = 7.0f;
        private static final long EXPLOSION_DURATION_MS = 300;

        private Animator getExplosionAnimator(String scaleXOrScaleY, ImageView explosionImage) {
            ObjectAnimator animator = ObjectAnimator.ofFloat(explosionImage, scaleXOrScaleY, MAXIMUM_XY_SCALE);
            animator.setInterpolator(new LinearInterpolator());
            animator.setDuration(EXPLOSION_DURATION_MS);
            return animator;
        }
    }

    public class BonkAnimatorListener implements AnimatorListener {

        @Override
        public void onAnimationCancel(Animator animation) {
        }

        @Override
        public void onAnimationEnd(Animator animation) {
        }

        @Override
        public void onAnimationRepeat(Animator animation) {
        }

        @Override
        public void onAnimationStart(Animator animation) {
        }

    }

    private class AxisImpactAnimatorListener extends BonkAnimatorListener {
        private final String axis;
        private final Bullet bullet;
        private final boolean gameObjectIsOwnedByThisUser;
        private Trajectory trajectoryAnimation;

        public AxisImpactAnimatorListener(Trajectory trajectoryAnimation, Bullet bullet, String axis,
                boolean gameObjectIsOwnedByThisUser) {
            this.axis = axis;
            this.bullet = bullet;
            this.gameObjectIsOwnedByThisUser = gameObjectIsOwnedByThisUser;
            this.trajectoryAnimation = trajectoryAnimation;
        }

        @Override
        public void onAnimationStart(Animator animation) {
            if (axis.equals("x")) {
                // x axis travel means impacting a Y (left or right) wall and mirroring the rotation across
                // the Y axis; i.e. R' = 180-R
                bullet.setRotation((float) (180.0 - bullet.getRotation()));
            }
            else {
                // y axis travel = top/bottom wall impact = mirror across the X axis (R' = -R)
                bullet.setRotation((float) (-bullet.getRotation()));
            }

            // check to see if this flight path will result in any collisions
            if (gameObjectIsOwnedByThisUser) {
                trajectoryAnimation.checkCurrentTrajectoryForImpactEvents(animation);
            }
        }
    }

    private class DisappearWhenFinishedAnimationListener extends BonkAnimatorListener {
        private final ImageView disappearingObject;

        public DisappearWhenFinishedAnimationListener(ImageView disappearingObject) {
            this.disappearingObject = disappearingObject;
        }

        @Override
        public void onAnimationEnd(Animator animation) {
            playgroundView.removeView(disappearingObject);
        }
    }
}