package lu.uni.mobile.bonk;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Random;

import org.json.JSONException;
import org.json.JSONObject;

import android.util.Log;
import lu.uni.mobile.bonk.GameObject.Block;
import lu.uni.mobile.bonk.GameObject.Bullet;
import lu.uni.mobile.bonk.GameObject.Player;
import lu.uni.mobile.bonk.GameObject.Target;

public class Game {
    private final PlayGameActivity activity;
    private final PlaygroundView playground;
    private int thisPlayerId;
    private Player thisPlayer; // the main player (i.e. the person holding the phone)
    private int nextObjectId;
    private int maxObjectId;
    private int blocksPerUser;
    private final long startTimeMilliseconds; // this is the END of the block placing and the START of the
    // game
    private static final Random random = new Random();

    private HashMap<Integer, GameObject> gameObjects = new HashMap<Integer, GameObject>();

    // these define the range for getXDouble(), getYDouble(), etc
    public static final double MAX_COORDINATE = 1.0;
    public static final double MIN_COORDINATE = 0.0;
    public static final long ROTATION_DURATION = 10;

    public final static int BIT_POSITION_OWNER_ID = 13;
    public final static int BIT_MASK_OBJECT_NUMBER = 0x1FFF; // 0001 1111 1111 1111
    public final static int BIT_MASK_OWNER_ID = 0xE000; // 1110 0000 0000 0000
    public static final byte IMPACT_MESSAGE_INDICATOR_BYTE = (byte) 255;

    public Game(PlayGameActivity activity, PlaygroundView playground, int userNumber, int blocksPerUser,
            long startTimeMilliseconds) {
        this.activity = activity;
        this.playground = playground;
        this.blocksPerUser = blocksPerUser;
        this.startTimeMilliseconds = startTimeMilliseconds;

        thisPlayerId = Game.getPlayerId(userNumber);
        nextObjectId = thisPlayerId + blocksPerUser + 1;
        maxObjectId = thisPlayerId + 255;
    }

    public void rotatePlayer(double rotateAmount) {
        thisPlayer.rotate(rotateAmount);
        sendObjectPositionsToOtherDevices();
    }

    /**
     * Any time gameObjects.put() or (GameObject).rotate() is called, this must be called to send a sync
     * message to the other devices
     */
    public void sendObjectPositionsToOtherDevices() {
        byte[] thisPlayerObjectStates = getEncodedObjectStates(thisPlayerId);
        activity.sendUdpByteString(thisPlayerObjectStates, thisPlayerObjectStates.length);
    }

    public Bullet fireBullet() {
        // bullet trajectory is perpendicular to player (+90 degrees; +64 in byte value)
        byte directionCcwByte = (byte) ((thisPlayer.getRotationCcwByte() + 64) % 256);
        int id = getNextObjectId();
        byte x = thisPlayer.getXByte();
        byte y = thisPlayer.getYByte();

        Bullet bullet = createOrUpdateBullet(id, x, y, directionCcwByte);
        sendObjectPositionsToOtherDevices();
        return bullet;
    }

    private Bullet createOrUpdateBullet(int id, byte x, byte y, byte directionCcwByte) {
        // this will also animate the bullet
        Bullet bullet = null;
        if (gameObjects.containsKey(id)) {
            bullet = (Bullet) gameObjects.get(id);
            bullet.setXByte(x);
            bullet.setYByte(y);
            bullet.setRotationByte(directionCcwByte);
            bullet.reAnimate();
        }
        else {
            bullet = new Bullet(activity, playground, id, x, y, directionCcwByte, this);
            playground.addView(bullet);
            gameObjects.put(id, bullet);
        }
        return bullet;
    }

    private int getNextObjectId() {
        if (nextObjectId == maxObjectId) {
            // loop around, possibly overwriting the oldest bullet in the field (but not likely)
            nextObjectId = thisPlayerId + blocksPerUser + 1;
        }
        else {
            nextObjectId++;
        }
        return nextObjectId;
    }

    /**
     * @return Negative number: during block-placing phase, the negative number of MS until game starts.
     *         Positive number: during game-play phase, number of elapsed MS in the game so far
     */
    public long getCurrentGameTimer() {
        return System.currentTimeMillis() - startTimeMilliseconds;
    }

    /**
     * Number of TENTHS of a second (not MS) since the start of the game
     * 
     * @return
     */
    public byte[] getCurrentGameTimerBytes() {
        long timeStamp = getCurrentGameTimer() / 100;
        if (timeStamp < 0) {
            // there is no way of returning both signed and unsigned. Besides, you shouldn't be calling this
            // method during the block-placing phase, which is when the result would be negative.
            return new byte[] { 0, 0 };
        }
        byte[] result = { (byte) (timeStamp >> 8), (byte) (timeStamp & 0xF) };
        return result;
    }

    public static int getPlayerId(int playerNumber) {
        return getObjectId(playerNumber, 0);
    }

    public static int getObjectId(int playerNumber, int objectNumber) {
        return playerNumber << BIT_POSITION_OWNER_ID + objectNumber;
    }

    public static int getObjectOwnerId(int objectId) {
        return objectId & BIT_MASK_OWNER_ID;
    }

    public static int getObjectNumber(int objectId) {
        return (objectId & BIT_MASK_OBJECT_NUMBER);
    }

    public static int getOwnerPlayerNumber(int objectId) {
        // int from 0-5; 0=system, 1-5=the players
        return objectId >> BIT_POSITION_OWNER_ID;
    }

    public GameObject getGameObject(int objectId) {
        return gameObjects.get(objectId);
    }

    public HashMap<Integer, GameObject> getGameObjects() {
        return gameObjects;
    }

    public int getNumberOfBlocksPerUser() {
        return blocksPerUser;
    }

    /**
     * Get a byte-encoded representation of all object states owned by the given playerId.
     */
    private byte[] getEncodedObjectStates(int playerId) {
        // big enough to store timestamp and all objects, if necessary
        byte[] result = new byte[gameObjects.size() * 5 + 2];

        // first two bytes = timeStamp
        byte[] timeStamp = getCurrentGameTimerBytes();
        result[0] = timeStamp[0];
        result[1] = timeStamp[1];

        // object data starts after the timeStamp
        int actualLength = 2;
        for (Integer objectId : gameObjects.keySet()) {
            if (getObjectOwnerId(objectId) == playerId) {
                byte[] objectBytes = gameObjects.get(objectId).getByteStringRepresentation();
                result[actualLength++] = objectBytes[0]; // ID
                result[actualLength++] = objectBytes[1]; // ID
                result[actualLength++] = objectBytes[2]; // x
                result[actualLength++] = objectBytes[3]; // y
                result[actualLength++] = objectBytes[4]; // direction
            }
        }
        return Arrays.copyOfRange(result, 0, actualLength);
    }

    public void handleObjectTrackingMessage(int timeStamp, int objectId, byte x, byte y, byte directionCcwByte) {
        // TODO don't ignore the timeStamp
        if (gameObjects.containsKey(objectId)) {
            // Log.e("NetworkCommunicator", String.format(
            // "Updating object: (%d, %d, %d, %d)",
            // objectId,
            // x,
            // y,
            // directionCcwByte));
            gameObjects.get(objectId).updateFromByteValues(x, y, directionCcwByte);
        }
        else if (getObjectOwnerId(objectId) == objectId) {
            // this is a new player object
            Log.e("NetworkCommunicator", String.format(
                "Skipping player creation (just seems weird, doesn't it?): (%d, %d, %d, %d)",
                objectId,
                x,
                y,
                directionCcwByte));
            // TODO is there any reason to follow these orders?
            // gameObject = new Player(activity, objectId, x, y);
        }
        else if (getObjectNumber(objectId) <= blocksPerUser) {
            // this is a new block object
            Log.e("NetworkCommunicator", String.format(
                "Skipping block object creation (just seems weird, doesn't it?): (%d, %d, %d, %d)",
                objectId,
                x,
                y,
                directionCcwByte));
            // TODO
            // gameObject = new Block(activity, objectId, x, y);
        }
        else {
            // this is a new bullet object
            Log.e("NetworkCommunicator", String.format(
                "Creating bullet: (%d, %d, %d, %d)",
                objectId,
                x,
                y,
                directionCcwByte));
            // TODO
            createOrUpdateBullet(objectId, x, y, directionCcwByte);
        }
    }

    public void handleImpactMessage(int timeStamp, int objectId1, int objectId2) {
        // TODO
        Log.e("NetworkCommunicator", String.format(
            "Skipping impact message: (%d, %d, %d)",
            timeStamp,
            objectId1,
            objectId2));
    }

    /**
     * Add a player to the field during game setup phase
     */
    public void initPlayer(int id, byte x, byte y, String name, String playerIpAddress) {
        if (!gameObjects.containsKey(id)) {
            Player player = new Player(activity, playground, id, x, y, name, playerIpAddress);
            playground.addView(player);
            gameObjects.put(player.id, player);
            if (id == thisPlayerId) {
                thisPlayer = player;
            }
        }
    }

    /**
     * Add a target to the field during game setup phase
     */
    public void initTarget(byte x, byte y) {
        if (!gameObjects.containsKey(0)) {
            Target target = new Target(activity, playground, x, y);
            playground.addView(target);
            gameObjects.put(target.id, target);
        }
    }

    /**
     * Add a block to the field during game setup phase
     */
    public void initBlock(int id, byte x, byte y) {
        if (!gameObjects.containsKey(id)) {
            Block block = new Block(activity, playground, id, x, y);
            playground.addView(block);
            gameObjects.put(id, block);
        }
    }

    /**
     * the player ID (object ID) of the person holding the device
     */
    public int getThisPlayerId() {
        return thisPlayerId;
    }

    public void createRandomGame() {
        // playground must be empty
        gameObjects.clear();
        nextObjectId = thisPlayerId + blocksPerUser + 1;

        try {
            JSONObject gameSetup = new JSONObject();
            gameSetup.put(
                NetworkCommunicator.MSG_TYPE_JSON_KEY,
                NetworkCommunicator.MSG_TYPE_PLAYGROUND_SETUP);
            gameSetup.put("target", getRandomXY());

            JSONObject players = new JSONObject();
            gameSetup.put("players", players);
            players.put(String.valueOf(Game.getPlayerId(1)), getRandomPlayer(Game.getPlayerId(1)));
            players.put(String.valueOf(Game.getPlayerId(2)), getRandomPlayer(Game.getPlayerId(2)));
            // players.put(String.valueOf(Game.getPlayerId(3)), getRandomPlayer(Game.getPlayerId(3)));

            byte[] data = gameSetup.toString().getBytes();
            activity.sendUdpByteString(data, data.length);
        } catch (JSONException e) {
            e.printStackTrace();
        }

    }

    private JSONObject getRandomXY() throws JSONException {
        JSONObject result = new JSONObject();
        result.put("x", random.nextInt(255));
        result.put("y", random.nextInt(255));
        return result;
    }

    private JSONObject getRandomPlayer(int playerId) throws JSONException {
        JSONObject result = getRandomXY();

        // TODO use the human-chosen names
        int playerNumber = Game.getOwnerPlayerNumber(playerId);
        result.put("name", "Player " + String.valueOf(playerNumber));
        if (playerId == 1) {
            result.put("ipAddress", activity.networkSender.getMyIpAddress().getHostAddress());
        }
        else {
            result.put("ipAddress", "10.0.2." + String.valueOf(99 + playerNumber));
        }
        JSONObject blocks = new JSONObject();
        result.put("blocks", blocks);
        for (int i = 1; i <= getNumberOfBlocksPerUser(); i++) {
            blocks.put(String.valueOf(playerId + i), getRandomXY());
        }
        return result;
    }
}
