package lu.uni.mobile.bonk;

import java.util.HashMap;

import android.util.Log;
import android.view.ViewGroup.LayoutParams;
import android.widget.ImageView;

public abstract class GameObject extends ImageView {

    public final int id;
    protected PlaygroundView playground;
    protected int playgroundWidth;
    protected int playgroundHeight;
    private double currentRotation = 0.0f;

    protected GameObject(PlayGameActivity activity, PlaygroundView playground, int id, byte x, byte y,
            byte directionCcwByte) {
        super(activity);
        this.id = id;
        this.playground = playground;

        playgroundHeight = playground.getHeight();
        playgroundWidth = playground.getWidth();

        setXByte(x);
        setYByte(y);
        setRotationByte(directionCcwByte);
        setImageResource(getResourceId());
    }

    public void updateFromByteValues(byte x, byte y, byte direction) {
        setXByte(x);
        setYByte(y);
        setRotationByte(direction);
    }

    public void setRotationByte(byte directionCcwByte) {
        int directionCcwInt = uInt(directionCcwByte);
        float directionCcwDegrees = 360.0f * ((float) directionCcwInt / 256.0f);
        setRotation(-directionCcwDegrees); // negative converts CCW to CW
    }

    public void setXByte(byte x) {
        int xInt = uInt(x);
        double xDouble = ((double) xInt) / 256.0;
        setXDouble(xDouble);
    }

    /**
     * get an unsigned int from a byte (Java casts as signed int by default)
     */
    private int uInt(byte byteValue) {
        return byteValue & 0xFF;
    }

    public void setXDouble(double x) {
        float xPixel = (float) (x * (playgroundWidth - getPixelWidth()));
        setX(xPixel);
        Log.i("ValueAnimator", String.format("%s\tX\t%.2f", this.getClass().getSimpleName(), x));
    }

    public byte getXByte() {
        double xDouble = getXDouble();
        double xByteAsDouble = Math.min(xDouble * 256.0, 255.0);
        return (byte) xByteAsDouble;
    }

    /**
     * @return x-axis location, from 0.0 (object is aligned to left edge) to 1.0 (object is aligned to right
     *         edge)
     */
    public double getXDouble() {
        return getX() / (playgroundWidth - getPixelWidth());
    }

    public void setYByte(byte y) {
        int yInt = uInt(y);
        double yDouble = ((double) yInt) / 256.0;
        setYDouble((double) yDouble);
    }

    public void setYDouble(double y) {
        float yPixel = (float) ((Game.MAX_COORDINATE - y) * (playgroundHeight - getPixelHeight()));
        setY(yPixel);
        Log.i("ValueAnimator", String.format("%s\tY\t%.2f", this.getClass().getSimpleName(), y));
    }

    public byte getYByte() {
        double yDouble = getYDouble();
        double yByteAsDouble = Math.min(yDouble * 256.0, 255.0);
        return (byte) yByteAsDouble;
    }

    /**
     * @return y-axis location, from 0.0 (object is aligned to bottom edge) to 1.0 (object is aligned to top
     *         edge)
     */
    public double getYDouble() {
        int height = playgroundHeight - getPixelHeight();
        return (height - getY()) / height;
    }

    public float getPixel(String axis, double coordinate) {
        if (axis == "x") {
            return getPixelX(coordinate);
        }
        if (axis == "y") {
            return getPixelY(coordinate);
        }
        return 0;
    }

    /**
     * Get the pixel value of the left edge of this object, if it were located at x on a 0.0-1.0 scale. This
     * calculation is useful for making sure objects don't go off the screen.
     */
    public float getPixelX(double x) {
        return (float) ((x / Game.MAX_COORDINATE) * (playgroundWidth - getPixelWidth()));
    }

    /**
     * Get the pixel value of the top edge of this object, if it were located at y on a 0.0-1.0 scale. This
     * calculation is useful for making sure objects don't go off the screen.
     */
    public float getPixelY(double y) {
        // convert units AND flip top & bottom
        return (float) (((Game.MAX_COORDINATE - y) / Game.MAX_COORDINATE) * (playgroundHeight - getPixelHeight()));
    }

    protected int getPixelHeight() {
        return (int) (0.08 * playgroundHeight);
    }

    protected int getPixelWidth() {
        return (int) (0.08 * playgroundWidth);
    }

    protected abstract int getResourceId();

    public byte[] getByteStringRepresentation() {
        // each object tracking message is a 5-byte string: objectID[2]|x|y|direction
        byte[] result = new byte[5];

        // add the 2-byte object ID.
        result[0] = (byte) ((getGameObjectId() & 0xFF00) >> 8);
        result[1] = (byte) (getGameObjectId() & 0xFF);

        // add the x & y bytes
        result[2] = getXByte();
        result[3] = getYByte();

        // add the direction byte
        result[4] = getRotationCcwByte();
        return result;
    }

    @Override
    public int hashCode() {
        return getGameObjectId();
    }

    public int getGameObjectId() {
        return id;
    }

    @Override
    public boolean equals(Object o) {
        if (o == null) {
            return false;
        }
        return this.hashCode() == o.hashCode();
    }

    public void rotate(double rotateAmount) {
        float newRotation = (float) (currentRotation + rotateAmount);
        // disabled. Using this causes the network update to fail, because it sends the old values.
        // ObjectAnimator animator = ObjectAnimator.ofFloat(this, "rotation", newRotation);
        // animator.setDuration(10);
        // animator.start();
        setRotation(newRotation);
    }

    @Override
    public void setRotation(float newRotation) {
        super.setRotation(newRotation);
        currentRotation = newRotation % 360.0;
    }

    @Override
    public float getRotation() {
        return (float) currentRotation;
    }

    public double getRotationCwDegrees() {
        return currentRotation;
    }

    public double getRotationCcwDegrees() {
        return -currentRotation;
    }

    public byte getRotationCcwByte() {
        // TODO test me
        return (byte) (int) (getRotationCcwDegrees() * 256.0F / 360.0F);
    }

    public long getRemainingLifeMS() {
        return -1; // -1 = immortal
    }

    public double getSpeedPerMS() {
        return 0.0; // most objects don't move
    }

    @Override
    protected void onLayout(boolean changed, int left, int top, int right, int bottom) {
        playgroundHeight = playground.getHeight();
        playgroundWidth = playground.getWidth();
        LayoutParams params = getLayoutParams();
        params.width = getPixelWidth();
        params.height = getPixelHeight();
        setLayoutParams(params);
    }

    @Override
    protected void onAttachedToWindow() {
        super.onAttachedToWindow();
        LayoutParams params = getLayoutParams();
        params.width = getPixelWidth();
        params.height = getPixelHeight();
        setLayoutParams(params);
    }

    /**
     * does this GameObject move, or is it fixed?
     */
    public boolean isMovable() {
        return false;
    }

    public static class Target extends GameObject {

        public Target(PlayGameActivity activity, PlaygroundView playground, byte x, byte y) {
            super(activity, playground, 0, x, y, (byte) 0);
        }

        protected int getResourceId() {
            return R.drawable.target;
        }
    }

    public static class Player extends GameObject {
        private static HashMap<String, Player> ipAddressMap = new HashMap<String, Player>();
        private static HashMap<String, Player> nameMap = new HashMap<String, Player>();
        // TODO these are used for identifying other players' communications
        private String name;
        private String ipAddress;

        public Player(PlayGameActivity activity, PlaygroundView playground, int id, byte x, byte y,
                String name, String ipAddress) {
            super(activity, playground, id, x, y, (byte) 0);
            this.name = name;
            this.ipAddress = ipAddress;
            ipAddressMap.put(ipAddress, this);
        }

        public static Player getByIpAddress(String ipAddress) {
            return ipAddressMap.get(ipAddress);
        }

        public static Player getByName(String name) {
            return nameMap.get(name);
        }

        @Override
        protected int getResourceId() {
            switch (Game.getOwnerPlayerNumber(id)) {
            case 1:
                return R.drawable.player1;
            case 2:
                return R.drawable.player2;
            case 3:
                return R.drawable.player3;
            }
            return R.drawable.player4;
        }
    }

    public static class Block extends GameObject {

        protected Block(PlayGameActivity activity, PlaygroundView playground, int id, byte x, byte y) {
            super(activity, playground, id, x, y, (byte) 0);
        }

        protected int getResourceId() {
            return R.drawable.block;
        }

    }

    public static class Bullet extends GameObject {

        private final static long LIFE_SPAN_MS = 3000L;
        private BonkAnimator animator;
        private long createdAt;
        private Game game;

        public Bullet(PlayGameActivity activity, PlaygroundView playground, int id, byte x, byte y,
                byte directionCcwByte, Game game) {
            super(activity, playground, id, x, y, directionCcwByte);
            createdAt = System.currentTimeMillis();
            this.game = game;
            reAnimate();
        }

        public void reAnimate() {
            if (animator != null) {
                animator.stop();
            }
            animator = new BonkAnimator.Trajectory(playground, this, game);
            animator.animate();
        }

        protected int getResourceId() {
            return R.drawable.missile;
        }

        protected int getPixelHeight() {
            return super.getPixelHeight() / 2;
        }

        protected int getPixelWidth() {
            return super.getPixelWidth() / 2;
        }

        public void updateFromByteValues(byte x, byte y, byte direction) {
            super.updateFromByteValues(x, y, direction);
            animator.animate();
        }

        public long getRemainingLifeMS() {
            long remainingLife = LIFE_SPAN_MS + System.currentTimeMillis() - createdAt;
            return Math.max(remainingLife, 0); // 0 = dead
        }

        public double getSpeedPerMS() {
            return 0.001;
        }

        /**
         * does this GameObject move, or is it fixed?
         */
        @Override
        public boolean isMovable() {
            return true;
        }
    }
}
